unit uCsv_DPXML_Converter;

interface

uses Classes, SysUtils, DB, DBClient;

// CSV to XML DATAPACKET bidirectionnal converter
// AW Version 0.1 2013 10 30

type
  TMyFieldTypes = (mftUnknown, mftString, mftBoolean, mftInteger, mftFloat,
    mftDateTime);

  TEncoding = (Iso8859_1, Utf8);

const
  ArFormatCSV: Array [TMyFieldTypes] of char = (#0, 'a', 'l', 'i', 'r', 'd');
  ArFieldFormatDPXML: Array [TMyFieldTypes] of string = ('string', 'string', 'boolean',
    'i4', 'r8', 'dateTime');
  ArFormatFT: Array [TMyFieldTypes] of TFieldType = (ftUnknown, ftString,
    ftBoolean, ftInteger, ftFloat, ftDateTime);
  K1 = ';DATAPACKET;METADATA;FIELDS;FIELD;PARAMS;ROWDATA;ROW;';

type
  TCsv_DPXML_Converter = class
  public
    slDataXML: TStringList;
    slDataCSV: TStringList;
    procedure StructureXML;
    procedure OrganizeData;
    constructor Create;
    destructor Destroy; override;
    function GetFieldType(const c1: char): TFieldType;
    function GeFieldTypesDPXML (const c1: char): string;
    function GetMyFieldTypeFromFieldType (ft1 : TFieldType) : TMyFieldTypes ;
    function TransformToDP(stSource: String; encoding: TEncoding)
      : TMemoryStream;
  end;

procedure LoadDP2ClientDataSet(stSrc: string; const CDS1: TClientDataSet);

var
  aConverter: TCsv_DPXML_Converter;

implementation

procedure LoadDP2ClientDataSet(stSrc: string; const CDS1: TClientDataSet);
var
  MS1: TMemoryStream;
begin
  with aConverter do
  begin
    MS1 := TransformToDP(stSrc, Iso8859_1);
    slDataXML.LoadFromStream(MS1);
    MS1.Position := 0;
    CDS1.LoadFromStream(MS1);
    MS1.Free;
  end;
end;

{ TCsvConversion }

constructor TCsv_DPXML_Converter.Create;
begin
  inherited Create;
  FormatSettings.DecimalSeparator := '.';
  slDataXML := TStringList.Create;
  slDataCSV := TStringList.Create;
end;

destructor TCsv_DPXML_Converter.Destroy;
begin
  slDataXML.Free;
  slDataCSV.Free;
  inherited Destroy
end;

function TCsv_DPXML_Converter.GetFieldType(const c1: char): TFieldType;
var
  v1: TMyFieldTypes;
begin
  result := ArFormatFT[mftUnknown];
  for v1 := low(TMyFieldTypes) to high(TMyFieldTypes) do
    if c1 = ArFormatCSV[v1] then
      result := ArFormatFT[v1];
end;

function TCsv_DPXML_Converter.GeFieldTypesDPXML (const c1: char): string;
var
  v1: TMyFieldTypes;
begin
  result := ArFieldFormatDPXML[mftUnknown];
  for v1 := low(TMyFieldTypes) to high(TMyFieldTypes) do
    if c1 = ArFormatCSV[v1] then
      result := ArFieldFormatDPXML[v1];
end;

function TCsv_DPXML_Converter.GetMyFieldTypeFromFieldType (ft1 : TFieldType) : TMyFieldTypes ;
var
  v1: TMyFieldTypes;
begin
  result := mftUnknown ;
  for v1 := low(TMyFieldTypes) to high(TMyFieldTypes) do
    if ft1 = ArFormatFT[v1] then
      result := v1 ;
end;



procedure TCsv_DPXML_Converter.OrganizeData ;
type
  TTag = (UnknownTag, OpeningTag, ClosingTag, AutonomousTag);
var
  sl2: TStringList;
  st1, st_Indent: String;
  i1: integer;
  Tag1: TTag;

  function f1(st1: String): String;
  var
    st2: String;
    i1: integer;
  begin
    result := st1;
    st2 := copy(st1, 2, pos(' ', st1) - 2);
    if st2 = '' then
      if pos('/>', st1) > 0 then
        st2 := copy(st1, 2, length(st1) - 3)
      else
        st2 := copy(st1, 2, length(st1) - 2);
    if pos(';' + st2 + ';', K1) > 0 then
    begin
      // st2 :=
      delete(st1, 1, length(st2) + 1);
      repeat
        st1 := trim(st1);
        i1 := pos('"', st1);
        if i1 = 0 then
          break;
        st2 := st2 + ',' + copy(st1, 1, i1);
        delete(st1, 1, i1);
        i1 := pos('"', st1);
        st2 := st2 + copy(st1, 1, i1);
        delete(st1, 1, i1);
      until i1 = 0;
      result := st2;;
    end;
  end;

begin
  sl2 := TStringList.Create;
  st_Indent := '';
  for i1 := 0 to (slDataXML.Count - 1) do
  begin
    st1 := trim(slDataXML[i1]);
    Tag1 := UnknownTag;
    if pos('</', st1) = 1 then
      Tag1 := ClosingTag
    else if pos('/>', st1) = length(st1) - 1 then
      Tag1 := AutonomousTag
    else if (pos('<', st1) = 1) and (pos('>', st1) = length(st1)) then
      Tag1 := OpeningTag;
    case Tag1 of
      OpeningTag:
        begin
          sl2.Add(st_Indent + f1(st1));
          st_Indent := st_Indent + ' ';
        end;
      ClosingTag:
        delete(st_Indent, 1, 1);
      AutonomousTag:
        sl2.Add(st_Indent + f1(st1));
    end;
  end;
  slDataXML.text := sl2.text;
  sl2.Free;
end;

procedure TCsv_DPXML_Converter.StructureXML;
var
  MS1, ms2: TMemoryStream;
  sl1: TStringList;
  c1: char;
  i1, i2: integer;
  etat: char;
  bATraiter: boolean;
  // 1 Tag not really sure
  // 2 Tag very not really sure
  // B Tag ,C Comment , T = Text ,R Nothing
  // 3 Ending Comment not really sure
  // 4 Ending Comment more sure

  procedure WriteData(cEtat: char; st1: string);
  var
    i1: integer;
    c1: char;
  begin
    if cEtat <> #0 then
      etat := cEtat;
    for i1 := 1 to length(st1) do
    begin
      c1 := st1[i1];
      ms2.write(c1, 1);
      i2 := i2 + 1;
    end;
    bATraiter := false;
  end;

begin
  MS1 := TMemoryStream.Create;
  slDataXML.SaveToStream(MS1);
  slDataXML.clear;
  ms2 := TMemoryStream.Create;
  MS1.Position := 0;
  ms2.Position := 0;
  etat := 'R';
  i2 := 0;
  repeat
    bATraiter := true;
    i1 := MS1.Read(c1, 1);
    if (c1 = '<') and (etat in ['R', 'T', 'C']) then
      WriteData('1', #13#10 + c1);
    if bATraiter and (c1 = '!') and (etat = '1') then
      WriteData('2', c1);
    if bATraiter and (c1 = '-') and (etat = '2') then
      WriteData('C', c1);
    if bATraiter and (c1 <> '!') and (etat = '1') then
      WriteData('B', c1);
    if bATraiter and (c1 <> '-') and (etat = '2') then
      WriteData('B', c1);
    if bATraiter and (c1 = '-') and (etat = 'C') then
      WriteData('3', c1);
    if bATraiter and (c1 = '-') and (etat = '3') then
      WriteData('4', c1);
    if bATraiter and (c1 = '>') and (etat in ['B', '4']) then
      WriteData('R', c1 + #13#10);
    if bATraiter and not((c1 in [#13, #10]) and (etat = 'B')) then
      WriteData(#0, c1);
  until i1 = 0;
  MS1.Free;
  ms2.size := i2;
  ms2.Position := 0;
  sl1 := TStringList.Create;
  sl1.LoadFromStream(ms2);
  for i1 := 0 to sl1.Count - 1 do
    if (pos('<', sl1[i1]) > 0) and (pos('>', sl1[i1]) > 0) then
      slDataXML.Add(sl1[i1]);
  sl1.Free;
  ms2.Free;
end;

function TCsv_DPXML_Converter.TransformToDP(stSource: String; encoding: TEncoding)
  : TMemoryStream;
var
  slCSV, slXML, slChamp: TStringList;

  procedure WriteField;
  var
    stField, stType, stLength, st1, st2: string;
    i1: integer;
    sl2: TStringList;
  begin
    sl2 := TStringList.Create;
    sl2.Delimiter := ';' ;
    sl2.DelimitedText :=  StringReplace(slCSV[0],' ','_',[rfReplaceAll]) ;
    sl2.text := StringReplace(sl2.text, '_', ' ', [rfReplaceAll]);
    sl2.text := StringReplace(sl2.text, ':', '=', [rfReplaceAll]);
    for i1 := 0 to sl2.Count - 1 do
    begin
      st1 := '<FIELD attrname="#1#" fieldtype="#2#" #3#/>';
      stField := sl2.Names[i1];
      st2 := sl2.Values[stField];
      stType := GeFieldTypesDPXML (st2[1]);
      stLength := '';
      if GetFieldType(st2[1]) = ftString then
      begin
        stLength := copy(st2, 2, length(st2) - 1);
        stLength := StringReplace('width="#3#" ', '#3#', stLength, []);
      end;
      st1 := StringReplace(st1, '#1#', stField, []);
      st1 := StringReplace(st1, '#2#', stType, []);
      st1 := StringReplace(st1, '#3#', stLength, []);
      slXML.Add(st1);
      slChamp.Add(stField);
    end;
    sl2.Free;
  end;

  procedure WriteRecord;
  var
    i1, i2: integer;
    sl2: TStringList;
    st1: string;
  begin
    sl2 := TStringList.Create;
    slXML.Add('<ROWDATA>');
    sl2.Delimiter := ';' ;
    for i1 := 1 to slCSV.Count - 1 do
    begin
      sl2.clear;
      sl2.DelimitedText :=  StringReplace(slCSV[i1],' ','_',[rfReplaceAll]) ;
      sl2.text := StringReplace(sl2.text, '_', ' ', [rfReplaceAll]);
      st1 := '<ROW ';
      for i2 := 0 to slChamp.Count - 1 do
        st1 := st1 + slChamp[i2] + '="' + sl2[i2] + '" ';
      slXML.Add(st1 + '/>');
    end;
    slXML.Add('</ROWDATA>');
    sl2.Free;
  end;

begin
  result := TMemoryStream.Create;
  slXML := TStringList.Create;
  slChamp := TStringList.Create;
  slCSV := TStringList.Create;
  slCSV.text := stSource;
  case encoding of
    Iso8859_1 :  slXML.add('<?xml version="1.0" encoding="ISO-8859-1"?>')
    else slXML.Add('<?xml version="1.0" encoding="UTF-8"?>');
  end;
  slXML.Add('<DATAPACKET Version="2.0">');
  slXML.Add('<METADATA><FIELDS>');
  WriteField;
  slXML.Add('</FIELDS></METADATA>');
  WriteRecord;
  slXML.Add('</DATAPACKET>');
  slCSV.Free;
  slXML.SaveToStream(result);
  result.Position := 0;
end;

initialization

aConverter := TCsv_DPXML_Converter.Create;

finalization

aConverter.Free;

end.

